/**
 * Copyright (c) 2009-2011 Zauber S.A. <http://www.zaubersoftware.com/>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.zauberlabs.commons.ws.security;

import java.util.List;
import java.util.Set;

import javax.xml.namespace.QName;
import javax.xml.soap.SOAPElement;
import javax.xml.soap.SOAPException;
import javax.xml.soap.SOAPHeader;
import javax.xml.soap.SOAPMessage;
import javax.xml.ws.BindingProvider;
import javax.xml.ws.handler.Handler;
import javax.xml.ws.handler.MessageContext;
import javax.xml.ws.handler.soap.SOAPHandler;
import javax.xml.ws.handler.soap.SOAPMessageContext;

/**
 * Simple implementation of request signing using <a href=
 * "http://www.oasis-open.org/committees/download.php/13392/wss-v1.1-spec-pr-UsernameTokenProfile-01.htm"
 * >username-token profile</a>
 * 
 * @author Gaston Ponti
 * @since Oct 20, 2011
 */
public class SecurityHandler implements SOAPHandler<SOAPMessageContext> {

    /** Namespace for the SOAP Envelope. */
    private static String SOAPENV_NAMESPACE = "http://schemas.xmlsoap.org/soap/envelope/";
    

    /** The prefix that will be used for the SOAP Envelope namespace. */
    private static String SOAPENV_PREFIX = "soapenv";

    /** Namespace for the WS-Security SOAP header elements. */
    private static String WSSE_NAMESPACE = "http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-secext-1.0.xsd";

    /** The prefix that will be used for the WS-Security namespace. */
    private static String WSSE_PREFIX = "wsse";

    /**
     * The WS-Security URI that specifies that the password will be transmitted
     * as plain text.
     */
    private static String WSSE_PASSWORD_TEXT = "http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-username-token-profile-1.0#PasswordText";

    private final Credential credential;

    /**
     * This method created an instance of the SOAPSimpleCredentials class and
     * adds it as a handler to the bindingProvider supplied.
     * 
     * @param bindingProvider
     *            The client stub to which the handler will be added. The most
     *            convenient way to obtain the required bindingProvvider is to
     *            call one of the getPort methods on the Service class for the
     *            Web service and then cast the returned object to a
     *            BindingProvider.
     *            
     * @param username
     *            The id and tenant name for the user. This is of the form
     *            systemid@tenant.
     * @param password
     *            The password for the system user.
     *            
     */
    @SuppressWarnings("unchecked")
    public static void sign(final BindingProvider bindingProvider, final String username, final String password) {
        List<Handler> handlerChain = bindingProvider.getBinding().getHandlerChain();
        handlerChain.add(new SecurityHandler(new BasicCredential(username, password)));
        bindingProvider.getBinding().setHandlerChain(handlerChain);
    }

    /**
     * Creates a {@link SecurityHandler} with a specified Credential
     * 
     * @param credential
     */
    public SecurityHandler(final Credential credential) {
        this.credential = credential;
    }

    /**
     * Returns null as this handler doesn't process any Headers, it just adds
     * one.
     */
    public final Set<QName> getHeaders() {
        return null;
    }

    /**
     * Adds WS-Security header to request messages.
     */
    public final boolean handleMessage(final SOAPMessageContext smc) {
        Boolean outboundProperty = (Boolean) smc.get(MessageContext.MESSAGE_OUTBOUND_PROPERTY);
        if (outboundProperty.booleanValue()) {
            addWSSecurityHeader(smc, credential.getUsername(), credential.getPassword());
        }
        return true;
    }

    /**
     * Returns true, no action is taken for faults messages.
     */
    public final boolean handleFault(final SOAPMessageContext smc) {
        return true;
    }

    public void close(final MessageContext messageContext) {
    }

    /**
     * Adds a WS-Security header containing a UsernameToken to a SOAP message.
     * 
     * @param smc
     *            The SOAPMessageContent to which the WS-Security header will be
     *            added.
     * @param username
     *            The WS-Security username.
     * @param password
     *            The WS-Security password.
     * @throws java.lang.RuntimeException
     *             This exception will be thrown if a SOAPException occurs when
     *             modifying the message.
     */
    private void addWSSecurityHeader(final SOAPMessageContext smc, final String username, final String password) {

        try {
            // Get the SOAP Header
            SOAPMessage message = smc.getMessage();
            SOAPHeader header = message.getSOAPHeader();
            if (header == null) {
                // Create header as it doesn't already exist
                message.getSOAPPart().getEnvelope().addHeader();
                header = message.getSOAPHeader();
            }

            SOAPElement heSecurity = header.addChildElement("Security", WSSE_PREFIX, WSSE_NAMESPACE);
            heSecurity.addAttribute(
                    message.getSOAPPart().getEnvelope()
                            .createName("mustUnderstand", SOAPENV_PREFIX, SOAPENV_NAMESPACE), "1");

            SOAPElement heUsernameToken = heSecurity.addChildElement("UsernameToken", WSSE_PREFIX,
                    WSSE_NAMESPACE);

            heUsernameToken.addChildElement("Username", WSSE_PREFIX, WSSE_NAMESPACE).addTextNode(username);

            SOAPElement hePassword = heUsernameToken.addChildElement("Password", WSSE_PREFIX, WSSE_NAMESPACE);

            hePassword.addAttribute(message.getSOAPPart().getEnvelope().createName("Type"), WSSE_PASSWORD_TEXT);
            hePassword.addTextNode(password);

        } catch (SOAPException e) {
            throw new RuntimeException("Failed to add WS-Security header to request", e);
        }
    }
}
